from swap_user.to_email_otp.base_models import AbstractEmailOTPUser


class EmailOTPUser(AbstractEmailOTPUser):
    """
    Concrete implementation of user with email login.
    """

    class Meta:
        swappable = "AUTH_USER_MODEL"
        verbose_name = "OTP email user"
        verbose_name_plural = "OTP email users"



from django.contrib.auth.base_user import AbstractBaseUser as DjangoAbstractBaseUser
from django.contrib.auth.models import PermissionsMixin
from django.db import models
from django.utils.translation import gettext_lazy as _

from phonenumber_field.modelfields import PhoneNumberField

from swap_user.helpers import check_password
from swap_user.to_phone_otp.managers import PhoneOTPUserManager


class AbstractPhoneOTPUser(PermissionsMixin):
    """
    Abstract PhoneUser implementation - subclass this class to provide your own
    custom User class with `phone` field.

    Main difference between this model and other models - that it doesn't
    have a `password` field.
    You need to use OTP (One Time Password) to authenticate user
    of this model and this require extra work from you on backend.

    Provides fields:
    - phone
    - is_active (required by django.contrib.admin)
    - is_staff (required by django.contrib.admin)

    Provides attributes:
    - USERNAME_FIELD
    - EMAIL_FIELD

    REQUIRED_FIELDS by default will include USERNAME_FIELD and password.
    """

    phone = PhoneNumberField(verbose_name=_("phone number"), unique=True,)
    is_active = models.BooleanField(
        verbose_name=_("active"),
        default=False,
        help_text=_(
            "Designates whether this user should be treated as active. "
            "Unselect this instead of deleting accounts."
        ),
    )
    is_staff = models.BooleanField(
        verbose_name=_("staff status"),
        default=False,
        help_text=_("Designates whether the user can log into this admin site."),
    )
    last_login = models.DateTimeField(_("last login"), blank=True, null=True)

    objects = PhoneOTPUserManager()

    REQUIRED_FIELDS = []
    USERNAME_FIELD = "phone"
    # Fix `django.contrib.auth.forms.PasswordResetForm`
    EMAIL_FIELD = "email"

    class Meta:
        abstract = True

    def __str__(self):
        return self.phone.as_e164

    def check_password(self, user_one_time_password: str) -> bool:
        """
        Method that checks for OTP passwords match.
        """

        user_id = str(self.id)

        is_valid = check_password(user_id, user_one_time_password)

        return is_valid

    clean = DjangoAbstractBaseUser.clean
    get_username = DjangoAbstractBaseUser.get_username
    natural_key = DjangoAbstractBaseUser.natural_key
    is_anonymous = DjangoAbstractBaseUser.is_anonymous
    is_authenticated = DjangoAbstractBaseUser.is_authenticated
    get_email_field_name = DjangoAbstractBaseUser.get_email_field_name
    normalize_username = DjangoAbstractBaseUser.normalize_username

    get_short_name = __str__
    get_full_name = __str__



from swap_user.to_email_otp.base_managers import BaseEmailOTPUserManager


class EmailOTPUserManager(BaseEmailOTPUserManager):
    """
    Concrete implementation of manager for EmailOTPUser.
    """

    pass



from django.contrib import admin

from swap_user.base_forms import BaseUserOptionalFieldsForm, BaseUserRequiredFieldsForm


class BaseUserAdmin(admin.ModelAdmin):
    add_form_class = BaseUserRequiredFieldsForm
    change_form_class = BaseUserOptionalFieldsForm

    def get_form(self, request, obj=None, change=False, **kwargs):
        if change:
            form = self.change_form_class
        else:
            form = self.add_form_class

        kwargs["form"] = form

        return super().get_form(request, obj, change, **kwargs)



from django import forms
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from swap_user.helpers import check_user_was_banned, get_banned_user_cache_key, normalize_username


UserModel = get_user_model()


class GetOTPForm(forms.ModelForm):
    """
    Unfortunately we can't check User exist or not at this screen by the
    security reasons - if we will show error when User doesn't exist,
    attacker can just check all the emails / phones.
    """

    class Meta:
        model = UserModel
        fields = (UserModel.USERNAME_FIELD,)

    def clean(self):
        """
        We are preventing unique validation by overriding this method and
        adding extra check for user ban.
        """

        username_field = UserModel.USERNAME_FIELD
        raw_username = self.cleaned_data[username_field]
        username = normalize_username(raw_username)

        self._check_user_is_banned(username)

        return self.cleaned_data

    def _check_user_is_banned(self, username: str):
        """
        We are banning user for too many invalid login
        attempts.
        Here we are checking for this.
        """

        cache_key = get_banned_user_cache_key(username)
        is_banned = check_user_was_banned(cache_key)

        if is_banned:
            message = _("You are banned - contact with admin.")
            code = "banned"
            raise forms.ValidationError(message, code)


class CheckOTPForm(GetOTPForm):
    """
    Here we are checking User presence among with OPT check.
    """

    otp = forms.CharField(label=_("OTP"), widget=forms.PasswordInput(attrs={"autofocus": True}))

    def clean(self):
        """
        Method that allows to us pass through few validation checks:
            - User DB presence
            - OTP password validity
        """

        username_field = UserModel.USERNAME_FIELD
        username = self.cleaned_data[username_field]
        otp = self.cleaned_data["otp"]

        self._check_user_is_banned(username)
        user = self._get_user(username)
        self._check_password(user, otp)

    def _get_user(self, username: str) -> UserModel:
        """
        Check User presence in our DB or not.
        """

        username_field = UserModel.USERNAME_FIELD
        query_data = {username_field: username}

        try:
            user = UserModel.objects.get(**query_data)
        except UserModel.DoesNotExist:
            message = _("Invalid credentials.")
            code = "invalid_credentials"
            raise forms.ValidationError(message, code)

        return user

    def _check_password(self, user: UserModel, otp: str):
        """
        Check backend cached OTP with user provided OTP.
        """

        if not user.check_password(otp):
            message = _("Invalid credentials.")
            code = "invalid_credentials"
            raise forms.ValidationError(message, code)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['swap_user',
 'swap_user.otp',
 'swap_user.tests',
 'swap_user.to_email',
 'swap_user.to_email.migrations',
 'swap_user.to_email_otp',
 'swap_user.to_email_otp.migrations',
 'swap_user.to_phone',
 'swap_user.to_phone.migrations',
 'swap_user.to_phone_otp',
 'swap_user.to_phone_otp.migrations']

package_data = \
{'': ['*'], 'swap_user.to_email_otp': ['templates/admin/*']}

install_requires = \
['django-phonenumber-field[phonenumbers]>=5.2.0,<6.0.0', 'django>=2.2']

setup_kwargs = {
    'name': 'django-swap-user',
    'version': '0.9.1',
    'description': '(Beta) Simple and flexible way to swap default Django User',
    'long_description': '# Django-Swap-User (Beta)\n\n## About\nIf you are tired from copying one custom user model from one project to another ones - use this package.\nThis will do all for you. \n\n\n## Installation\n```\npip install django-swap-user\n```\n\n## Basic usage\n1. Choose one of models that suits for you and copy related settings from the table:\n\n| Application name | Username field | Description                                                            | `INSTALLED_APPS`                               | `AUTH_USER_MODEL`                      | Replace `django.contrib.admin` to          |\n|------------------|----------------|------------------------------------------------------------------------|------------------------------------------------|----------------------------------------|--------------------------------------------|\n| `to_email`       | `email`        | User with `email` username                                             | ```"swap_user", "swap_user.to_email",```       | `"swap_to_email.EmailUser"`            | not required                               |                                  \n| `to_email_otp`   | `email`        | User with `email` username, without `password` and OPT authentication  | ```"swap_user", "swap_user.to_email_otp",```   | `"swap_to_email_otp.EmailOTPUser"`     | `"swap_user.to_email_otp.apps.SiteConfig"` | \n| `to_phone`       | `phone`        | User with `phone` username                                             | ```"swap_user", "swap_user.to_phone",```       | `"swap_to_phone.PhoneUser"`            | not required                               |                                            \n| `to_phone_otp`   | `phone`        | User with `phone` username, without `password`  and OTP authentication | ```"swap_user", "swap_user.to_phone_otp",```   | `"swap_to_phone_otp.PhoneOTPUser"`     | `"swap_user.to_phone_otp.apps.SiteConfig"` | \n\n2. Add corresponding app to `INSTALLED_APPS`:\n```python\nINSTALLED_APPS = [\n    ...\n    "swap_user",\n    "swap_user.to_email",\n    ...\n]\n```\n\n3. Change `AUTH_USER_MODEL` to corresponding:\n```python\nAUTH_USER_MODEL = "swap_to_email.EmailUser"\n```\n\n4. If required replace `django.contrib.admin` in `INSTALLED_APPS` to something.\n\n5. Apply migrations:\n```bash\npython manage.py migrate swap_to_email\n```\n\n\n## Architecture\nApplication `swap_user` split into 3 apps:\n  - `to_email` - provides user with `email` username field\n  - `to_email_otp` - provides user with `email` username field and OTP (One Time Password) authentication\n  - `to_phone` - provides user with `phone` username field\n  - `to_phone_otp` - provides user with `phone` username field and OTP (One Time Password) authentication\n  \n  \n## Why so unusual architecture?\nBecause if we leave them in one app, they all will create migrations and tables - such approach leads us to redundant tables.\nThey will be treated as 3 custom models within the same app, which causes perplexing and cognitive burden.\n\nWith such approach (when there is a common app which contains internal apps) - the user \nchoose and connect only the specific user model which suits best for concrete business-logic. \n\nI have found such approach at Django REST Framework `authtoken` application and decide to use it - reference is [here](https://github.com/encode/django-rest-framework/tree/master/rest_framework/authtoken).\n\n\n## Providing User model at start of project\nWhen you are starting a project from zero or scratch - this is a best moment to provide custom User model.\nBecause you have\'t a lot of migrations or you can easily regenerate them. Moreover, Django\'s [official docs](https://docs.djangoproject.com/en/dev/topics/auth/customizing/#using-a-custom-user-model-when-starting-a-project)\nrecommend to provide custom User model even if you are fully satisfied with default one - in future it will be easier to extend custom model that fits into your business cases.\n\n\n## Providing User model at mid of project\nThis is a harder way of doing things, but it is still possible to do:\n- Do all the steps at testing database and ONLY IF all of them was successful - try to apply at production environment\n- Please note that these steps are **common** - they fit in most cases, but in some circumstances you need act on situation.\n- Create a backup of your database\n- Add stable tag into your repository or save a commit hash reference\n- Pray to the heavens\n- Remove all of yours migrations in every app of Django\'s project\n- Remove all records from `django_migrations` table, for example with SQL `TRUNCATE django_migrations`\n- Now we have a "clean" state, so we can change default model\n- Generate new migrations for all of your applications - `python manage.py makemigrations` \n- Now we need to [fake migrate](https://docs.djangoproject.com/en/4.0/ref/django-admin/#cmdoption-migrate-fake) because we already have all the tables with data\n- First fake the `auth` application because we are depending from this one - `python manage.py migrate --fake auth`\n- Install this library, follow instructions and apply migrations\n- Then fake rest of migrations we have - `python manage.py migrate --fake`\n- Run your application!',
    'author': 'Artem Innokentiev',
    'author_email': 'artinnok@protonmail.com',
    'maintainer': 'Artem Innokentiev',
    'maintainer_email': 'artinnok@protonmail.com',
    'url': 'http://github.com/artinnok/django-swap-user',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from swap_user.to_email.base_models import AbstractEmailUser


class EmailUser(AbstractEmailUser):
    class Meta:
        swappable = "AUTH_USER_MODEL"
        verbose_name = "email user"
        verbose_name_plural = "email users"



import secrets
from typing import Union

from django.core.cache import cache

from swap_user.settings import swap_user_settings


DEFAULT_COUNTER_VALUE = 1
DEFAULT_BANNED_VALUE = False


def generate_otp() -> str:
    """
    Method that generates OTP (One Time Password).
    """

    otp_alphabet = swap_user_settings.OTP_ALPHABET
    otp_length = swap_user_settings.OTP_LENGTH

    otp_digits = [secrets.choice(otp_alphabet) for _ in range(otp_length)]
    otp = "".join(otp_digits)

    return otp


def set_key_to_cache(
    cache_key: str, value: Union[int, str], expire=swap_user_settings.OTP_TIMEOUT
) -> str:
    """
    Saves value into cache with provided key.
    """

    cache.set(cache_key, value, expire)

    return value


def get_otp_cache_key(user_id: str) -> str:
    """
    Generates cache key for storing OTP (One Time Password) per user.
    """

    cache_key = swap_user_settings.OTP_PATTERN.format(user_id=user_id)

    return cache_key


def get_invalid_login_cache_key(user_id: str) -> str:
    """
    Get invalid login cache key for concrete user.
    """

    cache_key = swap_user_settings.INVALID_LOGIN_PATTERN.format(user_id=user_id)

    return cache_key


def get_banned_user_cache_key(user_id: str) -> str:
    """
    Get banned user cache key
    """

    cache_key = swap_user_settings.BANNED_USER_PATTERN.format(user_id=user_id)

    return cache_key


def increase_counter_of_invalid_login(cache_key: str) -> int:
    """
    Increase counter of invalid login by 1.
    """

    try:
        current_counter = cache.incr(cache_key)
    except ValueError:
        current_counter = cache.set(cache_key, DEFAULT_COUNTER_VALUE)

    return current_counter


def check_password(user_id: str, user_one_time_password: str) -> bool:
    """
    Function that checks for OTP passwords match.
    """

    cache_key = get_otp_cache_key(user_id)
    backend_one_time_password = cache.get(cache_key)

    is_same_password = backend_one_time_password == user_one_time_password

    return is_same_password


def check_user_was_banned(cache_key: str) -> bool:
    """
    Checks user is banned or not.
    """

    is_banned = cache.get(cache_key, DEFAULT_BANNED_VALUE)

    return is_banned


def normalize_username(username: str) -> str:
    """
    Username normalization function:
        - Removes all spaces
        - Transforms to lowercase
    """

    normalized = username.strip().lower()

    return normalized
